import triad_openvr
import time
import sys
import struct
import socket

sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
server_address = ('10.0.1.48', 8051)

v = triad_openvr.triad_openvr()
v.print_discovered_objects()

if len(sys.argv) == 1:
    interval = 1/250
elif len(sys.argv) == 2:
    interval = 1/float(sys.argv[1])
else:
    print("Invalid number of arguments")
    interval = False
    
if interval:
    while(True):
        start = time.time()
        txt = ""
        data =  v.devices["tracker_1"].get_pose_quaternion()
        sent = sock.sendto(struct.pack('d'*len(data), *data), server_address)
        print("\r" + txt, end="")
        sleep_time = interval-(time.time()-start)
        if sleep_time>0:
            time.sleep(sleep_time)


from typing import Any

import grpc

import vive_tracker_apiserver.grpc.tracker_packet_pb2 as pb2
import vive_tracker_apiserver.grpc.tracker_packet_pb2_grpc as pb2_grpc


class ClientStream:
    response: Any = None

    def __init__(self, response: Any):
        self.response = response

    def read(self):
        return next(self.response)

    def close(self):
        self.response.cancel()


class Client:
    endpoint: str
    channel: grpc.Channel = None
    stub: pb2_grpc.TrackerServiceStub = None

    def __init__(self, endpoint: str) -> None:
        self.endpoint = endpoint
        self.channel = grpc.insecure_channel(self.endpoint)
        self.stub = pb2_grpc.TrackerServiceStub(self.channel)

    def reconnect(self):
        if self.channel is not None:
            self.channel.close()
        self.channel = grpc.insecure_channel(self.endpoint)
        self.stub = pb2_grpc.TrackerServiceStub(self.channel)

    def get_group(self):
        return self.stub.GetTrackerGroup(
            pb2.TrackerRequest(uid="")
        )

    def open_group_stream(self):
        return ClientStream(
            self.stub.GetTrackerGroupStream(
                pb2.TrackerRequest(uid="")
            )
        )

    def start_recording(self, tag: str):
        return self.stub.StartRecording(
            pb2.RecordingRequest(tag=tag)
        )

    def stop_recording(self):
        return self.stub.StopRecording(
            pb2.RecordingRequest(tag="")
        )


def get_client(endpoint: str) -> Client:
    return Client(endpoint)



import logging
from concurrent import futures
from typing import List, Dict, Any
import time

import grpc

import vive_tracker_apiserver.grpc.tracker_packet_pb2 as pb2
import vive_tracker_apiserver.grpc.tracker_packet_pb2_grpc as pb2_grpc
from vive_tracker_apiserver.apiserver.app import Application
from vive_tracker_apiserver.common.Config import Config

logger = logging.getLogger("apiserver.server")


def create_tracker_single_packet(meta: Dict[str, Any], pos_quat_7d: List[float] = None, uid: str = None):
    if uid is not None:
        meta['uid'] = uid
    if pos_quat_7d is None:
        meta['valid'] = False
        return pb2.TrackerSingleResponse(meta=pb2.TrackerMetaResponse(**meta))
    else:
        meta['valid'] = True
        return pb2.TrackerSingleResponse(meta=pb2.TrackerMetaResponse(**meta),
                                         pos_x=pos_quat_7d[0],
                                         pos_y=pos_quat_7d[1],
                                         pos_z=pos_quat_7d[2],
                                         rot_w=pos_quat_7d[3],
                                         rot_x=pos_quat_7d[4],
                                         rot_y=pos_quat_7d[5],
                                         rot_z=pos_quat_7d[6])


def create_tracker_group_packet(meta: Dict[str, Any], payload: Dict[str, List[float]]):
    if any([x is None for x in payload.values()]):
        meta['valid'] = False
    else:
        meta['valid'] = True

    return pb2.TrackerGroupResponse(meta=pb2.TrackerMetaResponse(**meta),
                                    payload=[
                                        create_tracker_single_packet(meta=meta, pos_quat_7d=v, uid=k) for k, v in payload.items()
                                    ])


class TrackerService(pb2_grpc.TrackerService):
    def __init__(self, config: Config) -> None:
        super().__init__()
        self.app: Application = Application(config)
        self.app.start_thread()
        self.app.logger.info(f"vive tracker service listen at {config.api_port}")
        self.app.logger.info(f"vive tracker config {config}")

    def GetTrackerGroup(self,
                        target,
                        options=(),
                        channel_credentials=None,
                        call_credentials=None,
                        insecure=False,
                        compression=None,
                        wait_for_ready=None,
                        timeout=None,
                        metadata=None):
        logger.debug(f"GetTrackerGroup: {target}")
        meta, payload = self.app.get_group()
        return create_tracker_group_packet(meta, payload)

    def GetTrackerGroupStream(self,
                              target,
                              options=(),
                              channel_credentials=None,
                              call_credentials=None,
                              insecure=False,
                              compression=None,
                              wait_for_ready=None,
                              timeout=None,
                              metadata=None):
        logger.info(f"GetTrackerGroupStream: {target}")
        options.add_callback(lambda: logger.info(
            "GetTrackerGroupStream: context deadline exceeded"))

        meta, payload = self.app.get_group()
        last_index = meta['index']
        yield create_tracker_group_packet(meta, payload)

        while True:
            start_t = time.time()
            desired_index = last_index + 1
            meta, payload = self.app.get_group(index=desired_index)

            if not meta['valid'] or meta['index'] < desired_index:
                sleep_t = self.app.interval_s - (time.time() - start_t)
                if sleep_t > 0:
                    time.sleep(sleep_t)
                continue

            else:
                if meta['index'] > desired_index:
                    logger.warning('slow client operation detected')
                last_index = meta['index']
                yield create_tracker_group_packet(meta, payload)

    def GetTrackerSingle(self,
                         target,
                         options=(),
                         channel_credentials=None,
                         call_credentials=None,
                         insecure=False,
                         compression=None,
                         wait_for_ready=None,
                         timeout=None,
                         metadata=None):
        logger.debug(f"GetTrackerSingle: {target}")
        meta, payload = self.app.get_single(target.addr)
        return create_tracker_single_packet(meta, payload)

    def GetTrackerSingleStream(self,
                               target,
                               options=(),
                               channel_credentials=None,
                               call_credentials=None,
                               insecure=False,
                               compression=None,
                               wait_for_ready=None,
                               timeout=None,
                               metadata=None):
        logger.info(f"GetTrackerSingleStream: {target}")
        options.add_callback(lambda: logger.info(
            "GetTrackerGroupStream: context deadline exceeded"))

        meta, payload = self.app.get_single(target.uid)
        last_index = meta['index']
        yield create_tracker_single_packet(meta, payload)

        while True:
            start_t = time.time()
            desired_index = last_index + 1
            meta, payload = self.app.get_single(target.uid, index=desired_index)
            if not meta['valid'] or meta['index'] < desired_index:
                sleep_t = self.app.interval_s - (time.time() - start_t)
                if sleep_t > 0:
                    time.sleep(sleep_t)
                continue

            else:
                if meta['index'] > desired_index:
                    logger.warning('slow client operation detected')
                last_index = meta['index']
                yield create_tracker_single_packet(meta, payload)

    def StartRecording(self,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        logger.info(f"StartRecording: {target}")
        err = self.app.start_recording(tag=target.tag)
        if err is not None:
            return pb2.RecordingResponse(started=False, error=str(err))
        else:
            return pb2.RecordingResponse(started=True, error="")


    def StopRecording(self,
                      target,
                      options=(),
                      channel_credentials=None,
                      call_credentials=None,
                      insecure=False,
                      compression=None,
                      wait_for_ready=None,
                      timeout=None,
                      metadata=None):
        logger.info(f"StopRecording: {target}")
        err = self.app.stop_recording()
        if err is not None:
            return pb2.RecordingResponse(started=False, error=str(err))
        else:
            return pb2.RecordingResponse(started=True, error="")

    # def GetFIFOStatus(self, request: pb2.ForceGetFIFOStatusRequest, context):
    #     return pb2.ForceStatusResponse(status=self.app.fifo_status)
    #
    # def SetFIFOStatus(self, request: pb2.ForceSetFIFOStatusRequest, context):
    #     logger.info(f"SetFIFOStatus: {request}")
    #     if request.status:
    #         err = self.app.start_fifo()
    #     else:
    #         err = self.app.stop_fifo()
    #     return pb2.ForceStatusResponse(status=True, err=str(err))
    #
    # def GetPacket(self, request: pb2.ForcePacketRequest, context):
    #     logger.info(f"GetForcePacket: {request}")
    #     data = self.app.get()
    #     return create_packet(data)
    #
    # def GetPacketStream(self, request: pb2.ForcePacketRequest, context):
    #     logger.info(f"GetForcePacketStream: {request}")
    #     context.add_callback(lambda: logger.info(
    #         "GetPacketStream: context deadline exceeded"))
    #     while True:
    #         data = self.app.get()
    #         # print(data)
    #         yield create_packet(data)
    #
    # def ResetPacketCache(self, request: pb2.ForcePacketRequest, context):
    #     logger.info(f"ResetForcePacketCache: {request}")
    #     self.app.force_data_queue.queue.clear()
    #     return pb2.ForceStatusResponse(status=True, err=str(None))
    #
    # def ToggleRecording(self, request: pb2.ForceToggleRecordingRequest, context):
    #     logger.info(f"ToggleRecording: {request}")
    #     if request.start:
    #         err = self.app.start_recording(request.tag)
    #     else:
    #         err = self.app.stop_recording()
    #     return pb2.ForceStatusResponse(status=True, err=str(err))


def get_server(cfg: Config):
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    pb2_grpc.add_TrackerServiceServicer_to_server(
        TrackerService(cfg), server)
    server.add_insecure_port(f'{cfg.api_interface}:{cfg.api_port}')
    return server


if __name__ == '__main__':
    server = get_server(Config('./config.yaml'))
    server.start()
    logging.basicConfig(level=logging.INFO)

    try:
        while True:
            time.sleep(86400)
    except KeyboardInterrupt:
        server.stop(0)



import numpy as np
from scipy.spatial.transform import Rotation as R
def fix_vive_pose_matrix(pose_matrix):
    if pose_matrix is None:
        return None
    else:
        OPENXR_TO_RHS = np.array([[1, 0, 0], [0, 0, -1], [0, 1, 0]])

        pose_matrix = np.array(
            [
                list(pose_matrix[0]),
                list(pose_matrix[1]),
                list(pose_matrix[2]),
                [0, 0, 0, 1]
            ]
        )

        pose_matrix[:3, 3:4] = OPENXR_TO_RHS @ pose_matrix[:3, 3:4]
        # pose_matrix[:3, 3:4] = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]]) @ pose_matrix[:3, 3:4]
        pose_matrix[:3, :3] = OPENXR_TO_RHS @ np.linalg.inv(np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]]) @ pose_matrix[:3, :3])

        return np.concatenate([pose_matrix[:3, 3], R.from_matrix(pose_matrix[:3, :3]).as_quat()])


import time
from vive_tracker_apiserver.client import Client

def get_description(data):
    return str([f'id:{item.meta.uid} x: {item.pos_x:.2f} y: {item.pos_y:.2f} z: {item.pos_z:.2f}' for item in data.payload])

def test():
    endpoint = "localhost:8080"

    c = Client(endpoint)

    response = c.start_recording("test" + str(time.time()))
    print("StartRecording client received: " + str(response))

    time.sleep(20)

    response = c.stop_recording()
    print("StopRecording client received: " + str(response))

if __name__ == '__main__':
    test()


import argparse
import logging
import time

from vive_tracker_apiserver.apiserver.server import get_server
from vive_tracker_apiserver.common import Config


def main(args):
    logging.basicConfig(level=logging.INFO)
    cfg = Config(args.config)
    if cfg.valid is False:
        logging.error("invalid config file")
        exit(1)

    server = get_server(cfg)
    server.start()

    try:
        while True:
            time.sleep(86400)
    except KeyboardInterrupt:
        server.stop(0)


def entry_point(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", type=str, default="./config.yaml")
    run_args = parser.parse_args(argv[1:])
    main(run_args)


if __name__ == '__main__':
    import sys

    exit(entry_point(sys.argv))



import time
from vive_tracker_apiserver.client import Client

def get_description(data):
    return str([f'id:{item.meta.uid} x: {item.pos_x:.2f} y: {item.pos_y:.2f} z: {item.pos_z:.2f}' for item in data.payload])

def test():
    endpoint = "localhost:8080"

    c = Client(endpoint)
    response = c.get_group()
    print("GetTrackerGroup client received: " + str(response))

    stream = c.open_group_stream()
    print("GetTrackerGroupStream client received: " + str(response))

    try:
        while True:
            data = stream.read()
            if data.meta.valid == False:
                time.sleep(0.05)
                continue
            print('\r'+get_description(data), end="")
    except KeyboardInterrupt as e:
        stream.close()

if __name__ == '__main__':
    test()


import pprint

from vive_tracker_apiserver.common import TrackerConfig
from vive_tracker_apiserver.third_party import triad_openvr as triad_openvr

def list_devices(args):
    v = triad_openvr.triad_openvr()
    trackers = [TrackerConfig(uid=x.get_serial(), name=name) for name, x in v.devices.items()]
    print("Detected Trackers:")
    print("-"*30)
    print("{:<20} {:<10}".format("UUID", "Name"))
    print("-"*30)
    for t in trackers:
        print("{:<20} {:<10}".format(t.uid, t.name))
    print("-"*30)


# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import vive_tracker_apiserver.grpc.tracker_packet_pb2 as tracker__packet__pb2


class TrackerServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetTrackerGroup = channel.unary_unary(
                '/protocol.TrackerService/GetTrackerGroup',
                request_serializer=tracker__packet__pb2.TrackerRequest.SerializeToString,
                response_deserializer=tracker__packet__pb2.TrackerGroupResponse.FromString,
                )
        self.GetTrackerGroupStream = channel.unary_stream(
                '/protocol.TrackerService/GetTrackerGroupStream',
                request_serializer=tracker__packet__pb2.TrackerRequest.SerializeToString,
                response_deserializer=tracker__packet__pb2.TrackerGroupResponse.FromString,
                )
        self.GetTrackerSingle = channel.unary_unary(
                '/protocol.TrackerService/GetTrackerSingle',
                request_serializer=tracker__packet__pb2.TrackerRequest.SerializeToString,
                response_deserializer=tracker__packet__pb2.TrackerSingleResponse.FromString,
                )
        self.GetTrackerSingleStream = channel.unary_stream(
                '/protocol.TrackerService/GetTrackerSingleStream',
                request_serializer=tracker__packet__pb2.TrackerRequest.SerializeToString,
                response_deserializer=tracker__packet__pb2.TrackerSingleResponse.FromString,
                )
        self.StartRecording = channel.unary_unary(
                '/protocol.TrackerService/StartRecording',
                request_serializer=tracker__packet__pb2.RecordingRequest.SerializeToString,
                response_deserializer=tracker__packet__pb2.RecordingResponse.FromString,
                )
        self.StopRecording = channel.unary_unary(
                '/protocol.TrackerService/StopRecording',
                request_serializer=tracker__packet__pb2.RecordingRequest.SerializeToString,
                response_deserializer=tracker__packet__pb2.RecordingResponse.FromString,
                )


class TrackerServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetTrackerGroup(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTrackerGroupStream(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTrackerSingle(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTrackerSingleStream(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StartRecording(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StopRecording(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TrackerServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetTrackerGroup': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTrackerGroup,
                    request_deserializer=tracker__packet__pb2.TrackerRequest.FromString,
                    response_serializer=tracker__packet__pb2.TrackerGroupResponse.SerializeToString,
            ),
            'GetTrackerGroupStream': grpc.unary_stream_rpc_method_handler(
                    servicer.GetTrackerGroupStream,
                    request_deserializer=tracker__packet__pb2.TrackerRequest.FromString,
                    response_serializer=tracker__packet__pb2.TrackerGroupResponse.SerializeToString,
            ),
            'GetTrackerSingle': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTrackerSingle,
                    request_deserializer=tracker__packet__pb2.TrackerRequest.FromString,
                    response_serializer=tracker__packet__pb2.TrackerSingleResponse.SerializeToString,
            ),
            'GetTrackerSingleStream': grpc.unary_stream_rpc_method_handler(
                    servicer.GetTrackerSingleStream,
                    request_deserializer=tracker__packet__pb2.TrackerRequest.FromString,
                    response_serializer=tracker__packet__pb2.TrackerSingleResponse.SerializeToString,
            ),
            'StartRecording': grpc.unary_unary_rpc_method_handler(
                    servicer.StartRecording,
                    request_deserializer=tracker__packet__pb2.RecordingRequest.FromString,
                    response_serializer=tracker__packet__pb2.RecordingResponse.SerializeToString,
            ),
            'StopRecording': grpc.unary_unary_rpc_method_handler(
                    servicer.StopRecording,
                    request_deserializer=tracker__packet__pb2.RecordingRequest.FromString,
                    response_serializer=tracker__packet__pb2.RecordingResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'protocol.TrackerService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class TrackerService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetTrackerGroup(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/protocol.TrackerService/GetTrackerGroup',
            tracker__packet__pb2.TrackerRequest.SerializeToString,
            tracker__packet__pb2.TrackerGroupResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTrackerGroupStream(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/protocol.TrackerService/GetTrackerGroupStream',
            tracker__packet__pb2.TrackerRequest.SerializeToString,
            tracker__packet__pb2.TrackerGroupResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTrackerSingle(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/protocol.TrackerService/GetTrackerSingle',
            tracker__packet__pb2.TrackerRequest.SerializeToString,
            tracker__packet__pb2.TrackerSingleResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTrackerSingleStream(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/protocol.TrackerService/GetTrackerSingleStream',
            tracker__packet__pb2.TrackerRequest.SerializeToString,
            tracker__packet__pb2.TrackerSingleResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StartRecording(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/protocol.TrackerService/StartRecording',
            tracker__packet__pb2.RecordingRequest.SerializeToString,
            tracker__packet__pb2.RecordingResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StopRecording(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/protocol.TrackerService/StopRecording',
            tracker__packet__pb2.RecordingRequest.SerializeToString,
            tracker__packet__pb2.RecordingResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import time
import sys
import openvr
import math
import json

from functools import lru_cache

# Function to print out text but instead of starting a new line it will overwrite the existing line
def update_text(txt):
    sys.stdout.write('\r'+txt)
    sys.stdout.flush()

#Convert the standard 3x4 position/rotation matrix to a x,y,z location and the appropriate Euler angles (in degrees)
def convert_to_euler(pose_mat):
    yaw = 180 / math.pi * math.atan2(pose_mat[1][0], pose_mat[0][0])
    pitch = 180 / math.pi * math.atan2(pose_mat[2][0], pose_mat[0][0])
    roll = 180 / math.pi * math.atan2(pose_mat[2][1], pose_mat[2][2])
    x = pose_mat[0][3]
    y = pose_mat[1][3]
    z = pose_mat[2][3]
    return [x,y,z,yaw,pitch,roll]

#Convert the standard 3x4 position/rotation matrix to a x,y,z location and the appropriate Quaternion
def convert_to_quaternion(pose_mat):
    # Per issue #2, adding a abs() so that sqrt only results in real numbers
    r_w = math.sqrt(abs(1+pose_mat[0][0]+pose_mat[1][1]+pose_mat[2][2]))/2
    r_x = (pose_mat[2][1]-pose_mat[1][2])/(4*r_w)
    r_y = (pose_mat[0][2]-pose_mat[2][0])/(4*r_w)
    r_z = (pose_mat[1][0]-pose_mat[0][1])/(4*r_w)

    x = pose_mat[0][3]
    y = pose_mat[1][3]
    z = pose_mat[2][3]
    return [x,y,z,r_w,r_x,r_y,r_z]

#Define a class to make it easy to append pose matricies and convert to both Euler and Quaternion for plotting
class pose_sample_buffer():
    def __init__(self):
        self.i = 0
        self.index = []
        self.time = []
        self.x = []
        self.y = []
        self.z = []
        self.yaw = []
        self.pitch = []
        self.roll = []
        self.r_w = []
        self.r_x = []
        self.r_y = []
        self.r_z = []

    def append(self,pose_mat,t):
        self.time.append(t)
        self.x.append(pose_mat[0][3])
        self.y.append(pose_mat[1][3])
        self.z.append(pose_mat[2][3])
        self.yaw.append(180 / math.pi * math.atan(pose_mat[1][0] /pose_mat[0][0]))
        self.pitch.append(180 / math.pi * math.atan(-1 * pose_mat[2][0] / math.sqrt(pow(pose_mat[2][1], 2) + math.pow(pose_mat[2][2], 2))))
        self.roll.append(180 / math.pi * math.atan(pose_mat[2][1] /pose_mat[2][2]))
        r_w = math.sqrt(abs(1+pose_mat[0][0]+pose_mat[1][1]+pose_mat[2][2]))/2
        self.r_w.append(r_w)
        self.r_x.append((pose_mat[2][1]-pose_mat[1][2])/(4*r_w))
        self.r_y.append((pose_mat[0][2]-pose_mat[2][0])/(4*r_w))
        self.r_z.append((pose_mat[1][0]-pose_mat[0][1])/(4*r_w))

def get_pose(vr_obj):
    return vr_obj.getDeviceToAbsoluteTrackingPose(openvr.TrackingUniverseStanding, 0, openvr.k_unMaxTrackedDeviceCount)


class vr_tracked_device():
    def __init__(self,vr_obj,index,device_class):
        self.device_class = device_class
        self.index = index
        self.vr = vr_obj

    @lru_cache(maxsize=None)
    def get_serial(self):
        return self.vr.getStringTrackedDeviceProperty(self.index, openvr.Prop_SerialNumber_String)

    def get_model(self):
        return self.vr.getStringTrackedDeviceProperty(self.index, openvr.Prop_ModelNumber_String)

    def get_battery_percent(self):
        return self.vr.getFloatTrackedDeviceProperty(self.index, openvr.Prop_DeviceBatteryPercentage_Float)

    def is_charging(self):
        return self.vr.getBoolTrackedDeviceProperty(self.index, openvr.Prop_DeviceIsCharging_Bool)


    def sample(self,num_samples,sample_rate):
        interval = 1/sample_rate
        rtn = pose_sample_buffer()
        sample_start = time.time()
        for i in range(num_samples):
            start = time.time()
            pose = get_pose(self.vr)
            rtn.append(pose[self.index].mDeviceToAbsoluteTracking,time.time()-sample_start)
            sleep_time = interval- (time.time()-start)
            if sleep_time>0:
                time.sleep(sleep_time)
        return rtn

    def get_pose_euler(self, pose=None):
        if pose == None:
            pose = get_pose(self.vr)
        if pose[self.index].bPoseIsValid:
            return convert_to_euler(pose[self.index].mDeviceToAbsoluteTracking)
        else:
            return None

    def get_pose_matrix(self, pose=None):
        if pose == None:
            pose = get_pose(self.vr)
        if pose[self.index].bPoseIsValid:
            return pose[self.index].mDeviceToAbsoluteTracking
        else:
            return None

    def get_velocity(self, pose=None):
        if pose == None:
            pose = get_pose(self.vr)
        if pose[self.index].bPoseIsValid:
            return pose[self.index].vVelocity
        else:
            return None

    def get_angular_velocity(self, pose=None):
        if pose == None:
            pose = get_pose(self.vr)
        if pose[self.index].bPoseIsValid:
            return pose[self.index].vAngularVelocity
        else:
            return None

    def get_pose_quaternion(self, pose=None):
        if pose == None:
            pose = get_pose(self.vr)
        if pose[self.index].bPoseIsValid:
            return convert_to_quaternion(pose[self.index].mDeviceToAbsoluteTracking)
        else:
            return None

    def controller_state_to_dict(self, pControllerState):
        # This function is graciously borrowed from https://gist.github.com/awesomebytes/75daab3adb62b331f21ecf3a03b3ab46
        # docs: https://github.com/ValveSoftware/openvr/wiki/IVRSystem::GetControllerState
        d = {}
        d['unPacketNum'] = pControllerState.unPacketNum
        # on trigger .y is always 0.0 says the docs
        d['trigger'] = pControllerState.rAxis[1].x
        # 0.0 on trigger is fully released
        # -1.0 to 1.0 on joystick and trackpads
        d['trackpad_x'] = pControllerState.rAxis[0].x
        d['trackpad_y'] = pControllerState.rAxis[0].y
        # These are published and always 0.0
        # for i in range(2, 5):
        #     d['unknowns_' + str(i) + '_x'] = pControllerState.rAxis[i].x
        #     d['unknowns_' + str(i) + '_y'] = pControllerState.rAxis[i].y
        d['ulButtonPressed'] = pControllerState.ulButtonPressed
        d['ulButtonTouched'] = pControllerState.ulButtonTouched
        # To make easier to understand what is going on
        # Second bit marks menu button
        d['menu_button'] = bool(pControllerState.ulButtonPressed >> 1 & 1)
        # 32 bit marks trackpad
        d['trackpad_pressed'] = bool(pControllerState.ulButtonPressed >> 32 & 1)
        d['trackpad_touched'] = bool(pControllerState.ulButtonTouched >> 32 & 1)
        # third bit marks grip button
        d['grip_button'] = bool(pControllerState.ulButtonPressed >> 2 & 1)
        # System button can't be read, if you press it
        # the controllers stop reporting
        return d

    def get_controller_inputs(self):
        result, state = self.vr.getControllerState(self.index)
        return self.controller_state_to_dict(state)

    def trigger_haptic_pulse(self, duration_micros=1000, axis_id=0):
        """
        Causes devices with haptic feedback to vibrate for a short time.
        """
        self.vr.triggerHapticPulse(self.index ,axis_id, duration_micros)

class vr_tracking_reference(vr_tracked_device):
    def get_mode(self):
        return self.vr.getStringTrackedDeviceProperty(self.index,openvr.Prop_ModeLabel_String).decode('utf-8').upper()
    def sample(self,num_samples,sample_rate):
        print("Warning: Tracking References do not move, sample isn't much use...")

class triad_openvr():
    def __init__(self, configfile_path=None):
        # Initialize OpenVR in the
        self.vr = openvr.init(openvr.VRApplication_Other)
        self.vrsystem = openvr.VRSystem()

        # Initializing object to hold indexes for various tracked objects
        self.object_names = {"Tracking Reference":[],"HMD":[],"Controller":[],"Tracker":[]}
        self.devices = {}
        self.device_index_map = {}
        poses = self.vr.getDeviceToAbsoluteTrackingPose(openvr.TrackingUniverseStanding, 0,
                                                               openvr.k_unMaxTrackedDeviceCount)

        # Loading config file
        if configfile_path:
            try:
                with open(configfile_path, 'r') as json_data:
                    config = json.load(json_data)
            except EnvironmentError: # parent of IOError, OSError *and* WindowsError where available
                print('config.json not found.')
                exit(1)

            # Iterate through the pose list to find the active devices and determine their type
            for i in range(openvr.k_unMaxTrackedDeviceCount):
                if poses[i].bDeviceIsConnected:
                    device_serial = self.vr.getStringTrackedDeviceProperty(i,openvr.Prop_SerialNumber_String).decode('utf-8')
                    for device in config['devices']:
                        if device_serial == device['serial']:
                            device_name = device['name']
                            self.object_names[device['type']].append(device_name)
                            self.devices[device_name] = vr_tracked_device(self.vr,i,device['type'])
        else:
            # Iterate through the pose list to find the active devices and determine their type
            for i in range(openvr.k_unMaxTrackedDeviceCount):
                if poses[i].bDeviceIsConnected:
                    self.add_tracked_device(i)

    def __del__(self):
        openvr.shutdown()

    def get_pose(self):
        return get_pose(self.vr)

    def poll_vr_events(self):
        """
        Used to poll VR events and find any new tracked devices or ones that are no longer tracked.
        """
        event = openvr.VREvent_t()
        while self.vrsystem.pollNextEvent(event):
            if event.eventType == openvr.VREvent_TrackedDeviceActivated:
                self.add_tracked_device(event.trackedDeviceIndex)
            elif event.eventType == openvr.VREvent_TrackedDeviceDeactivated:
                #If we were already tracking this device, quit tracking it.
                if event.trackedDeviceIndex in self.device_index_map:
                    self.remove_tracked_device(event.trackedDeviceIndex)

    def add_tracked_device(self, tracked_device_index):
        i = tracked_device_index
        device_class = self.vr.getTrackedDeviceClass(i)
        if (device_class == openvr.TrackedDeviceClass_Controller):
            device_name = "controller_"+str(len(self.object_names["Controller"])+1)
            self.object_names["Controller"].append(device_name)
            self.devices[device_name] = vr_tracked_device(self.vr,i,"Controller")
            self.device_index_map[i] = device_name
        elif (device_class == openvr.TrackedDeviceClass_HMD):
            device_name = "hmd_"+str(len(self.object_names["HMD"])+1)
            self.object_names["HMD"].append(device_name)
            self.devices[device_name] = vr_tracked_device(self.vr,i,"HMD")
            self.device_index_map[i] = device_name
        elif (device_class == openvr.TrackedDeviceClass_GenericTracker):
            device_name = "tracker_"+str(len(self.object_names["Tracker"])+1)
            self.object_names["Tracker"].append(device_name)
            self.devices[device_name] = vr_tracked_device(self.vr,i,"Tracker")
            self.device_index_map[i] = device_name
        elif (device_class == openvr.TrackedDeviceClass_TrackingReference):
            device_name = "tracking_reference_"+str(len(self.object_names["Tracking Reference"])+1)
            self.object_names["Tracking Reference"].append(device_name)
            self.devices[device_name] = vr_tracking_reference(self.vr,i,"Tracking Reference")
            self.device_index_map[i] = device_name

    def remove_tracked_device(self, tracked_device_index):
        if tracked_device_index in self.device_index_map:
            device_name = self.device_index_map[tracked_device_index]
            self.object_names[self.devices[device_name].device_class].remove(device_name)
            del self.device_index_map[tracked_device_index]
            del self.devices[device_name]
        else:
            raise Exception("Tracked device index {} not valid. Not removing.".format(tracked_device_index))

    def rename_device(self,old_device_name,new_device_name):
        self.devices[new_device_name] = self.devices.pop(old_device_name)
        for i in range(len(self.object_names[self.devices[new_device_name].device_class])):
            if self.object_names[self.devices[new_device_name].device_class][i] == old_device_name:
                self.object_names[self.devices[new_device_name].device_class][i] = new_device_name

    def print_discovered_objects(self):
        for device_type in self.object_names:
            plural = device_type
            if len(self.object_names[device_type])!=1:
                plural+="s"
            print("Found "+str(len(self.object_names[device_type]))+" "+plural)
            for device in self.object_names[device_type]:
                if device_type == "Tracking Reference":
                    print("  "+device+" ("+self.devices[device].get_serial()+
                          ", Mode "+self.devices[device].get_model()+
                          ", "+self.devices[device].get_model()+
                          ")")
                else:
                    print("  "+device+" ("+self.devices[device].get_serial()+
                          ", "+self.devices[device].get_model()+")")



import argparse
import sys

import vive_tracker_apiserver.apiserver as apiserver
import vive_tracker_apiserver.cmd as cmd

parser = argparse.ArgumentParser()

args = sys.argv[1:]
if len(args) == 0:
    exit(print("No arguments provided"))
if args[0] == "configure":
    exit(cmd.configure(args[1:]))
elif args[0] == "list":
    exit(cmd.list_devices(args[1:]))
elif args[0] == "serve":
    exit(apiserver.serve(args[1:]))
elif args[0] == "test.server":
    exit(cmd.test_server())
elif args[0] == "test.visualize":
    exit(cmd.test_visualize())
else:
    print("Unknown command {}, available commands are:".format(args[0]))
    print("-"*60)
    print("{:<20} {:<40}".format("Name", "Description"))
    print("-"*60)
    print("{:<20} {:<40}".format("configure", "Generate configuration"))
    print("{:<20} {:<40}".format("list", "List trackers"))
    print("{:<20} {:<40}".format("serve", "Launch the server"))
    print("{:<20} {:<40}".format("test.server", "Test server connection"))
    print("{:<20} {:<40}".format("test.visualize", "Test visualization with Open3d"))
    print("-"*60)


import time

import tqdm
from py_cli_interaction import must_parse_cli_string

from vive_tracker_apiserver.client import Client


def get_description(data):
    return str([f'id:{item.meta.uid} x: {item.pos_x:.2f} y: {item.pos_y:.2f} z: {item.pos_z:.2f}' for item in data.payload])


def test_server():
    endpoint = must_parse_cli_string("endpoint", "localhost:8080")

    c = Client(endpoint)
    response = c.get_group()
    print("GetTrackerGroup client received: " + str(response))

    stream = c.open_group_stream()
    print("GetTrackerGroupStream client received: " + str(response))

    try:
        with tqdm.tqdm() as pbar:
            while True:
                data = stream.read()
                if data.meta.valid == False:
                    time.sleep(0.05)
                    continue
                pbar.set_description(get_description(data))
                pbar.update(1)
                # time.sleep(0.05)
                # print)
    except KeyboardInterrupt as e:
        stream.close()


if __name__ == '__main__':
    test_server()



import logging
import os
import time

import numpy as np
import open3d as o3d
import transforms3d as t3d
from py_cli_interaction import must_parse_cli_string

from vive_tracker_apiserver.client import Client


def get_live_visualizer():
    vis = o3d.visualization.Visualizer()
    vis.create_window()
    return vis


def test_visualize():
    o3d.utility.set_verbosity_level(o3d.utility.VerbosityLevel.Debug)
    endpoint = must_parse_cli_string("endpoint", "localhost:8080")

    c = Client(endpoint)
    response = c.get_group()
    base = o3d.geometry.TriangleMesh.create_coordinate_frame(size=1, origin=np.array([0., 0., 0.]))
    tracker_meshes = {
        x.meta.uid: o3d.geometry.TriangleMesh.create_coordinate_frame(size=0.2, origin=np.array([0., 0., 0.])) for x in response.payload
    }
    tracker_transforms = {
        x.meta.uid: np.eye(4) for x in response.payload
    }
    vis = get_live_visualizer()
    [vis.add_geometry(source) for _, source in tracker_meshes.items()]
    vis.add_geometry(base)

    try:
        while True:
            data = c.get_group()
            if data.meta.valid == False:
                time.sleep(0.0005)
                continue

            for x in data.payload:
                T = np.eye(4)
                T[:3, :3] = t3d.quaternions.quat2mat([x.rot_w, x.rot_x, x.rot_y, x.rot_z])
                T[:3, 3] = [x.pos_x, x.pos_y, x.pos_z]
                T_diff = T @ np.linalg.inv(tracker_transforms[x.meta.uid])
                tracker_transforms[x.meta.uid] = T_diff @ tracker_transforms[x.meta.uid]

                tracker_meshes[x.meta.uid].transform(T_diff)
                vis.update_geometry(tracker_meshes[x.meta.uid])
            if not vis.poll_events():
                break
            vis.update_renderer()
    except KeyboardInterrupt as e:
        logging.info("got KeyboardInterrupt")

    finally:
        vis.destroy_window()
        vis.close()
        os._exit(0)


if __name__ == '__main__':
    test_visualize()



import pprint

import yaml
from py_cli_interaction import must_parse_cli_int, must_parse_cli_bool, must_parse_cli_string

from vive_tracker_apiserver.common import Config, TrackerConfig
from vive_tracker_apiserver.third_party import triad_openvr as triad_openvr


def main(args):
    try:
        v = triad_openvr.triad_openvr()
        while True:
            cfg = Config()

            cfg.api_port = must_parse_cli_int("Enter a port number", 1024, 65535, 8080)
            cfg.api_interface = must_parse_cli_string("Enter a interface", "0.0.0.0")
            cfg.debug = must_parse_cli_bool("Debug mode", False)
            cfg.trackers = [TrackerConfig(uid=x.get_serial(), name=name) for name, x in v.devices.items()]

            res = cfg.to_dict()
            print("Your configuration is:")
            pprint.pprint(res)
            confirm = must_parse_cli_bool("Confirm?", True)

            if confirm:
                break
            else:
                continue

        dest = must_parse_cli_string("Enter save destination", "./config.yaml")
        with open(dest, 'w') as f:
            yaml.dump({'vive_tracker_apiserver': res}, f)
    except KeyboardInterrupt as e:
        print("Exiting...")

    except EOFError as e:
        print("Exiting...")

def entry_point(argv):
    main(None)


if __name__ == '__main__':
    import sys

    main(sys.argv)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tracker_packet.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14tracker_packet.proto\x12\x08protocol\"\x1d\n\x0eTrackerRequest\x12\x0b\n\x03uid\x18\x01 \x01(\t\"\x1f\n\x10RecordingRequest\x12\x0b\n\x03tag\x18\x01 \x01(\t\"3\n\x11RecordingResponse\x12\x0f\n\x07started\x18\x01 \x01(\x08\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"S\n\x13TrackerMetaResponse\x12\r\n\x05valid\x18\x01 \x01(\x08\x12\r\n\x05index\x18\x02 \x01(\x04\x12\x11\n\tsys_ts_ns\x18\x03 \x01(\x04\x12\x0b\n\x03uid\x18\x04 \x01(\t\"\xad\x01\n\x15TrackerSingleResponse\x12+\n\x04meta\x18\x01 \x01(\x0b\x32\x1d.protocol.TrackerMetaResponse\x12\r\n\x05pos_x\x18\x02 \x01(\x01\x12\r\n\x05pos_y\x18\x03 \x01(\x01\x12\r\n\x05pos_z\x18\x04 \x01(\x01\x12\r\n\x05rot_w\x18\x05 \x01(\x01\x12\r\n\x05rot_x\x18\x06 \x01(\x01\x12\r\n\x05rot_y\x18\x07 \x01(\x01\x12\r\n\x05rot_z\x18\x08 \x01(\x01\"u\n\x14TrackerGroupResponse\x12+\n\x04meta\x18\x01 \x01(\x0b\x32\x1d.protocol.TrackerMetaResponse\x12\x30\n\x07payload\x18\x02 \x03(\x0b\x32\x1f.protocol.TrackerSingleResponse2\xf9\x03\n\x0eTrackerService\x12M\n\x0fGetTrackerGroup\x12\x18.protocol.TrackerRequest\x1a\x1e.protocol.TrackerGroupResponse\"\x00\x12U\n\x15GetTrackerGroupStream\x12\x18.protocol.TrackerRequest\x1a\x1e.protocol.TrackerGroupResponse\"\x00\x30\x01\x12O\n\x10GetTrackerSingle\x12\x18.protocol.TrackerRequest\x1a\x1f.protocol.TrackerSingleResponse\"\x00\x12W\n\x16GetTrackerSingleStream\x12\x18.protocol.TrackerRequest\x1a\x1f.protocol.TrackerSingleResponse\"\x00\x30\x01\x12K\n\x0eStartRecording\x12\x1a.protocol.RecordingRequest\x1a\x1b.protocol.RecordingResponse\"\x00\x12J\n\rStopRecording\x12\x1a.protocol.RecordingRequest\x1a\x1b.protocol.RecordingResponse\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tracker_packet_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TRACKERREQUEST._serialized_start=34
  _TRACKERREQUEST._serialized_end=63
  _RECORDINGREQUEST._serialized_start=65
  _RECORDINGREQUEST._serialized_end=96
  _RECORDINGRESPONSE._serialized_start=98
  _RECORDINGRESPONSE._serialized_end=149
  _TRACKERMETARESPONSE._serialized_start=151
  _TRACKERMETARESPONSE._serialized_end=234
  _TRACKERSINGLERESPONSE._serialized_start=237
  _TRACKERSINGLERESPONSE._serialized_end=410
  _TRACKERGROUPRESPONSE._serialized_start=412
  _TRACKERGROUPRESPONSE._serialized_end=529
  _TRACKERSERVICE._serialized_start=532
  _TRACKERSERVICE._serialized_end=1037
# @@protoc_insertion_point(module_scope)



from dataclasses import dataclass, field
from typing import Optional, List, Dict, Any

import yaml


@dataclass
class TrackerConfig:
    uid: str
    name: str = field(default='')


@dataclass
class Config:
    path: str = field(default='')
    trackers: List[TrackerConfig] = field(default_factory=list)
    api_port: int = field(default=8080)
    api_interface: str = field(default='0.0.0.0')
    debug: bool = field(default=False)
    data_path: str = field(default='./tracker_data', init=False)
    valid: bool = field(default=False, init=False)

    def __post_init__(self) -> None:
        err = self.load()
        if err is None:
            self.valid = True
        else:
            self.valid = False

    def from_dict(self, cfg: Dict[str, Any]):
        self.trackers = [TrackerConfig(**s) for s in cfg['trackers']]
        if 'api' in cfg.keys():
            self.api_port = cfg['api']['port'] if 'port' in cfg['api'] else self.api_port
            self.api_interface = cfg['api']['interface'] if 'interface' in cfg['api'] else self.api_interface
        self.debug = cfg['debug'] if 'debug' in cfg.keys() else self.debug
        self.data_path = cfg['data_path'] if 'data_path' in cfg else self.data_path
        return self

    def to_dict(self) -> Dict[str, Any]:
        return {
            "trackers": [
                {
                    "uid": s.uid,
                    'name': s.name
                }
                for s in self.trackers
            ],
            "api": {
                "port": self.api_port,
                "interface": self.api_interface
            },
            "data_path": self.data_path,
            "debug": self.debug
        }

    def load(self) -> Optional[Exception]:
        if self.path is not None:
            try:
                cfg_dict = yaml.load(open(self.path, "r"),
                                     Loader=yaml.SafeLoader)
            except Exception as e:
                return e

            try:
                self.from_dict(cfg_dict['vive_tracker_apiserver'])
                return None

            except Exception as e:
                return e

        else:
            return Exception("Config path is not set")

    def dump(self) -> Optional[Exception]:
        if self.path is not None:
            try:
                with open(self.path, 'w') as f:
                    yaml.dump(self.to_dict(), f)
                    return None
            except Exception as e:
                return Exception("Failed to dump config")
        else:
            return Exception("Config path is not set")



import argparse
import io
import json
import logging
import os
import threading
import time
import os.path as osp
from typing import Optional, Dict, Union, Tuple, Any, List

import numpy as np
from vive_tracker_apiserver.common.functional import fix_vive_pose_matrix

from vive_tracker_apiserver.common import Config
from vive_tracker_apiserver.third_party import triad_openvr


class Application:
    __CONFIG_BUFFER_SIZE__ = 1024
    logger: logging.Logger
    option: Config
    cnt: int
    interval_s: float
    tracker_state_buffer: List[Tuple[dict, dict]]
    vr: Optional[triad_openvr.triad_openvr] = None
    poll_thread: Optional[threading.Thread]
    update_thread: Optional[threading.Thread]
    recording_thread: Optional[threading.Thread]
    recording_cancel_ev: threading.Event

    def __init__(self, cfg) -> None:
        if isinstance(cfg, Config):
            self.option = cfg
        elif isinstance(cfg, str):
            self.option = Config(cfg)
        elif isinstance(cfg, argparse.Namespace):
            self.option = Config(cfg.config)
        else:
            raise TypeError(
                "cfg must be Config, str, or argparse.Namespace"
            )
        if self.option.valid is False:
            raise ValueError("invalid config file")

        self.logger = logging.getLogger("apiserver.app")
        self.vr = triad_openvr.triad_openvr()

        self.logger.info('loading openvr components')
        [(time.sleep(1), print(".", end="")) for _ in range(3)]
        print('\n')
        self.logger.info("server is tarted!")

        self.vr_device_mapping = {x.get_serial(): index for index, x in self.vr.devices.items() if x.get_serial() in [x.uid for x in self.option.trackers]}
        self.cnt = 0
        self.interval_s = 1 / 250
        self.tracker_state_buffer = [
                                        (
                                            dict(
                                                index=-1,
                                                sys_ts_ns=0,
                                                uid="",
                                                valid=False
                                            ),
                                            dict()
                                        )
                                    ] * self.__CONFIG_BUFFER_SIZE__
        self.update_thread = None
        self.recording_thread = None
        self.recording_cancel_ev = threading.Event()

    def start_thread(self):

        def poll_thread(interval_s: float):
            while True:
                self.vr.poll_vr_events()
                time.sleep(interval_s)

        def update_vive_tracker_thread(interval_s):
            while True:
                start_t = time.time()
                meta = dict(
                    index=self.cnt,
                    sys_ts_ns=time.time_ns(),
                    uid="",
                    valid=True
                )
                try:
                    payload = {device_uid: fix_vive_pose_matrix(self.vr.devices[self.vr_device_mapping[device_uid]].get_pose_matrix()) for device_uid in self.vr_device_mapping.keys()}
                    meta['sys_ts_ns'] = int((time.time_ns() + meta['sys_ts_ns']) / 2)
                    meta['valid'] = all(map(lambda x: x is not None, payload.values()))
                    self.tracker_state_buffer[self.cnt % self.__CONFIG_BUFFER_SIZE__] = (
                        meta,
                        payload
                    )

                    self.cnt += 1
                    sleep_t = interval_s - (time.time() - start_t)
                    if sleep_t > 0:
                        time.sleep(sleep_t)
                except Exception as e:
                    self.logger.error(e)
                    self.logger.error('update thread is terminated')
                    time.sleep(1)

        self.poll_thread = threading.Thread(target=poll_thread, args=(30,), daemon=True)
        self.poll_thread.start()
        self.update_thread = threading.Thread(target=update_vive_tracker_thread, args=(self.interval_s,), daemon=True)
        self.update_thread.start()

    def get_single(self, device_uid: str, index=None) -> Tuple[Dict[str, Any], Optional[List[Union[int, float]]]]:
        index = self.cnt - 1 if index is None else index
        meta, data = self.tracker_state_buffer[index % self.__CONFIG_BUFFER_SIZE__]
        meta['uid'] = device_uid
        return meta, data[device_uid]

    def get_group(self, index=None) -> Tuple[Dict[str, Any], Optional[Dict[str, List[Union[int, float]]]]]:
        index = self.cnt - 1 if index is None else index
        return self.tracker_state_buffer[index % self.__CONFIG_BUFFER_SIZE__]

    def start_recording(self, tag: str) -> Optional[Exception]:
        if self.recording_thread is not None:
            return Exception("already recording")
        self.recording_cancel_ev.clear()

        def dump_buffer_thread(file: io.FileIO, cancel_ev: threading.Event):
            meta, payload = self.get_group()
            last_index = meta['index']

            while True:
                if cancel_ev.is_set():
                    file.flush()
                    break
                start_t = time.time()
                desired_index = last_index + 1
                meta, payload = self.get_group(index=desired_index)

                if not meta['valid'] or meta['index'] < desired_index:
                    sleep_t = self.interval_s - (time.time() - start_t)
                    if sleep_t > 0:
                        time.sleep(sleep_t)
                    continue

                else:
                    if meta['index'] > desired_index:
                        self.logger.warning('slow client operation detected')
                    last_index = meta['index']
                    file.write('[' + json.dumps(meta) + ',' + json.dumps({k: v.tolist() if v is not None else v for k,v in payload.items()}) + ']\n')

        if not osp.exists(self.option.data_path):
            os.makedirs(self.option.data_path)
        recording_file = open(osp.join(self.option.data_path, f'{tag}.jsonl'), 'w')
        self.recording_thread = threading.Thread(target=dump_buffer_thread, args=(recording_file, self.recording_cancel_ev), daemon=True)
        self.recording_thread.start()

        return None

    def stop_recording(self) -> Optional[Exception]:
        if self.recording_thread is None:
            return Exception("not recording")
        self.recording_cancel_ev.set()
        try:
            self.recording_thread.join(timeout=5)
            self.recording_thread = None
            return None
        except Exception as e:
            return e

    def shutdown(self):
        return None


if __name__ == '__main__':
    import sys

    parser = argparse.ArgumentParser()
    parser.add_argument("--config", type=str, default="./config.yaml")
    run_args = parser.parse_args(sys.argv[1:])

    logging.basicConfig(level=logging.INFO)

    app = Application(run_args)

    try:
        while True:
            print(app.get_group())

    except KeyboardInterrupt as e:
        app.shutdown()



from vive_tracker_apiserver.third_party import triad_openvr as triad_openvr
import time
import sys

v = triad_openvr.triad_openvr()
v.print_discovered_objects()

if len(sys.argv) == 1:
    interval = 1/250
elif len(sys.argv) == 2:
    interval = 1/float(sys.argv[1])
else:
    print("Invalid number of arguments")
    interval = False

print(list(v.devices.keys()))
if interval:
    while(True):
        start = time.time()
        txt = ""

        res = v.devices["tracker_1"].get_pose_euler()
        if res is not None:
            for each in v.devices["tracker_1"].get_pose_euler():
                txt += "%.4f" % each
                txt += " "
        print("\r" + txt, end="")
        sleep_time = interval-(time.time()-start)
        if sleep_time>0:
            time.sleep(sleep_time)


import os

from setuptools import setup, find_packages

requires = open("requirements.txt", "r").readlines() if os.path.exists("requirements.txt") else open("./vive_tracker_apiserver.egg-info/requires.txt", "r").readlines()
print("#-------------------    ", str(os.listdir("./")))
setup(
    name="vive-tracker-apiserver",
    version="0.0.7",
    author="davidliyutong",
    author_email="davidliyutong@sjtu.edu.cn",
    description="Vive Tracker APIServer",
    packages=find_packages() + ["vive_tracker_apiserver.third_party.triad_openvr." + pkg for pkg in find_packages('vive_tracker_apiserver.third_party.triad_openvr')] + [
        'vive_tracker_apiserver.third_party.triad_openvr'],
    python_requires=">=3.7",
    install_requires=requires,
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type="text/markdown",
    include_package_data=True
)
